import argparse
import inspect
import types as python_types

try:
    from itertools import izip_longest as zip_longest

except ImportError:
    from itertools import zip_longest

try:
    from inspect import getfullargspec as getargspec

except ImportError:
    from inspect import getargspec

from .util import (argnamefmt,
                   getopt,
                   get_common_type,
                   get_description_and_epilog)

from .argtypes import TupleType, switch_off, switch_on, translate_types

def argmap(func, arguments=None):
    """Expose a function as a commandline tool.
    
    Additional annotations can be supplied by using the following decorators
    on the function:
     
     - :py:func:`argmap.arg` add type and help for a single argument
     - :py:func:`argmap.kwargs` add additional keyword arguments passed to the var kwargs
     - :py:func:`argmap.types` add type info for arguments
     - :py:func:`argmap.help` add a help text for arguments
     
    To use argmap, write a function and expose it to the commandline as::
        
        #!/usr/bin/env python
        # filenam: add.py
        
        @argmap.arg("a", int)
        @argmap.arg("b", int)
        def add(a, b):
            return a + b
        
        if __name__ == "__main__":
            print argmap.argmap(add)
    
    Now the script can be called as::
    
        python add.py 1 2
    
    :param function func: the function being annotated
    :param arguments: the command-line arguments, if not given arguments defaults 
                      to `sys.argv[1:]`
    :type arguments: list or None
    :return: the return value of the function being annotated

    .. seealso::
    
        See the section on :ref:`argument patterns<argument_patterns>` for a 
        list of common argument patterns and how they translate to command-line
        arguments.
    
    """
    parser = ArgumentParser(func)
    args, kwargs = parser.parse_args(arguments)
    return func(*args, **kwargs)

class ArgumentParser(object):
    """The actual argument parser used by argmap. 
    
    This class is a wrapper around argparse.ArgumentParser and builds the 
    correct parser from the function signature and the supplied annotations. 
    """
    def __init__(self, func):
        self.parser = self._new_argument_parser(func)
        self.arg_spec = {}
        
        self._init_func_info(func)
        self._init_annotation_info(func)
        
        self._build_parser()
    
    def _new_argument_parser(self, func):
        description, epilog = get_description_and_epilog(func)
        return argparse.ArgumentParser(description=description,
                                       epilog=epilog) 
    
    def _init_func_info(self, func):
        spec = getargspec(func)
        args, varargs, keywords, defaults = spec[:4]
        
        self.args = args
        self.varargs = varargs
        self.keywords = keywords
        self.defaults = defaults if (defaults is not None) else []
        
        self._remove_first_argument_of_methods(func)
    
    def _init_annotation_info(self, func):
        # the keyword arguments for add_argument for each argument
        self.kwargs = getopt(func, "kwargs")
        self.help = getopt(func, "help")
        self.types = getopt(func, "types")
        self.types = translate_types(self.types)

        self.post_types = {}
    
    def _remove_first_argument_of_methods(self, func):
        if isinstance(func, python_types.MethodType):
            self.args = self.args[1:]
            self.defaults = self.defaults[:len(self.args)]
    
    def hasedefault(self, idx):
        """Return whether the positional argument at the index has a default.
        """
        if idx is None:
            return False
        
        return idx >= (len(self.args) - len(self.defaults))
    
    def default(self, idx):
        """Return the default argument for the given index
        """
        # the index where the defaults start
        idxdiff = len(self.args) - len(self.defaults)
        
        if idx >= idxdiff:
            return self.defaults[idx - idxdiff]
        else:
            raise IndexError("argument {0} has no default".format(idx)) 
    
    def _get_arg(self, ns, arg, idx=None):
        """Retrieve the argument from the namespace and apply postprocessing.
        """
        val = getattr(ns, arg)
        
        # test whether we have a list type
        if isinstance(self.types.get(arg), list):
            # if so, test for default value, if nothing was added return default
            if self.hasedefault(idx) and val == []:
                val = self.default(idx)  
        
        type = self.post_types.get(arg, lambda x: x)
        return type(val)
    
    def parse_args(self, args=None):
        """ Parse the commandline.
        
        Returns the argument list and keyword argument dictionary. 
        """
        ns = self.parser.parse_args(args=args)
        
        args = [self._get_arg(ns, arg, idx) 
                for (idx, arg) in enumerate(self.args)]
        
        if self.varargs is not None:
            args.extend(self._get_arg(ns, self.varargs))
        
        kwargs = {arg: self._get_arg(ns, arg) for arg in self.kwargs.keys()}
        
        return args, kwargs
    
    def _set(self, arg, key, value):
        self.arg_spec.setdefault(arg, {}).setdefault(key, value)
    
    def _build_parser(self):
        self._add_types()
        self._add_help()
        self._parser_add_arguments()
    
    def _add_types(self):
        # inject any types
        for (arg, type) in self.types.items():
            # is this a post-type?
            if isinstance(type, TupleType):
                self._add_tuple_arg(arg, type)
                
            elif isinstance(type, list):
                self._add_list_arg(arg, type)
                
            elif isinstance(type, set):
                self._add_set_arg(arg, type)
                
            elif (type == switch_on) or (type == switch_off):
                self._add_switch_arg(arg, type)
                
            else:
                # else just add the type
                self._set(arg, "type", type)
    
    def _add_tuple_arg(self, arg, type):
        if self.varargs == arg: 
            # handle tuple types for varargs specially
            self.post_types[arg] = lambda l: [type(v) for v in 
                                              grouper(len(type), l)]
        
        else:
            self.post_types[arg] = type 
            self._set(arg, "nargs", len(type))
    
    def _add_list_arg(self, arg, type):
        if self.varargs == arg:
            raise ValueError("cannot annotate a vararg by a list-type")
        
        # list types have to be options
        self._set(arg, "name", "--" + argnamefmt(arg))
        self._set(arg, "action", "append")
        
        if isinstance(type[0], TupleType):
            self.post_types[arg] = lambda v: [type[0](i) for i in v]
            self._set(arg, "nargs", len(type[0]))
            
        else:
            self._set(arg, "type", type[0])
        
        # for a list override any default value by an empty list,
        # as further arguments are appended to the default
        # This is corrected in self._get_arg later on
        self._set(arg, "default", [])   
    
    def _add_set_arg(self, arg, type):
        type = list(type)

        common_type = get_common_type(type)

        self._set(arg, "type", common_type)
        self._set(arg, "choices", type)
    
    def _add_switch_arg(self, arg, type):
        # switch_* args have to be an option 
        self._set(arg, "name", "--" + argnamefmt(arg))
        self._set(arg, "action", ("store_true" if type == switch_on 
                                  else "store_false"))
        
    
    def _add_help(self):
        for (arg, help) in self.help.items():
            self._set(arg, "help", help)
    
    def _parser_add_arguments(self):
        # keep an ordered list of arguments added (correct order of pos args)
        args = []
        
        for (idx, arg) in enumerate(self.args):
            args.append(arg)
            argname = argnamefmt(arg)
            
            # if the argument has a default value, we add an option 
            if self.hasedefault(idx):
                self._set(arg, "name", "--" + argname)
                self._set(arg, "default", self.default(idx))
                
            else:
                # inconsistency in argparse:
                # names are not translated for pos args
                self._set(arg, "name", arg)
            
        if self.varargs is not None:
            # inconsistency in argparse:
            # names are not translated for pos args
            args.append(self.varargs)
            self._set(self.varargs, "name", self.varargs)
            self._set(self.varargs, "nargs", "*")
            
        for (arg, default) in self.kwargs.items():
            args.append(arg)
            self._set(arg, "name", "--" + argnamefmt(arg))
            self._set(arg, "default", default)
        
        for arg in args:
            spec = dict(self.arg_spec[arg])
            argname = spec.pop("name")
            self.parser.add_argument(argname, **spec)


def grouper(n, iterable, fillvalue=None):
    """grouper(3, 'ABCDEFG', 'x') --> ABC DEF Gxx
    
    Taken from the python docs
    """
    # NOTE: all arguments use the same iterator 
    args = [iter(iterable)] * n
    return zip_longest(fillvalue=fillvalue, *args)
