"""Specific variants of the tests using python3 syntax
"""
from unittest import  TestCase
import argmap

class TestSimple3(TestCase):
    def test_posargs_order_annotate(self):
        def func(a : str, b : str):
            return [a, b]

        self.assertListEqual(argmap.argmap(func, ["hello", "world"]),
                             ["hello", "world"])

    def test_types(self):
        def func(a : int, *b : float):
            return dict(a=a, b=list(b))

        res = argmap.argmap(func, ["1", "2.5", "3.5"])

        self.assertSetEqual(set(res.keys()), {"a", "b"})
        self.assertEqual(res["a"], 1)
        self.assertSequenceEqual(res["b"], [2.5, 3.5])

    def test_bool(self):
        def func(a : bool):
            return a

        self.assertEqual(argmap.argmap(func, ["yes"]), True)
        self.assertEqual(argmap.argmap(func, ["no"]), False)

    def test_switch_onoff(self):
        def func(a : True, b : False):
            return a, b

        self.assertTupleEqual(argmap.argmap(func, []), (False, True))
        self.assertTupleEqual(argmap.argmap(func, ["--a"]), (True, True))
        self.assertTupleEqual(argmap.argmap(func, ["--b"]), (False, False))
        self.assertTupleEqual(argmap.argmap(func, ["--b", "--a"]), (True, False))

class TestCompound3(TestCase):
    def test_tuples(self):
        def func(pos : (float,float)):
            return pos

        self.assertTupleEqual(argmap.argmap(func, ["2.0", "3.0"]), (2.0, 3.0))

    def test_list(self):
        def func(a : [int] =[]):
            return a

        self.assertSequenceEqual(argmap.argmap(func, ["--a", "1", "--a", "2"]),
                                 [1,2])

        self.assertSequenceEqual(argmap.argmap(func, []), [])

    def test_list_default(self):
        def func(a : [int] = [2,3]):
            return list(a)

        self.assertSequenceEqual(argmap.argmap(func, []), [2,3])
        self.assertSequenceEqual(argmap.argmap(func, ["--a", "1"]), [1])

    def test_list_of_tuples(self):
        def func(it : [(str,int)] = []):
            return it

        self.assertListEqual(argmap.argmap(func, ["--it", "a", "1",
                                                  "--it", "b", "2"]),
                             [("a", 1), ("b", 2)])

    def test_none_or_tuple(self):
        def func(a : argmap.none_or((str, int)) = None):
            return a

        self.assertIsNone(argmap.argmap(func, []))
        self.assertTupleEqual(argmap.argmap(func, ["--a", "b", "20"]), ("b", 20))

    def test_tuple_of_none_or_int_correct(self):
        def func(a : (argmap.none_or(int), argmap.none_or(int)) =(None,None)):
            return a

        self.assertEqual(argmap.argmap(func, []), (None, None))
        self.assertEqual(argmap.argmap(func, ["--a", "20", "30"]), (20, 30))

    def test_tuple_of_none_or_int_failing(self):
        """Incorrectly annotated, default arguments can break argument parsing.

        See test_tuple_of_none_or_int_correct for the correct annotation.
        """
        def func(a : (int, int) =(None,None)):
            return a

        with self.assertRaises(TypeError):
            argmap.argmap(func, [])


class TestChoices3(TestCase):
    def test_choice_pos_arg(self):
        def func(choice : {1,2,3}):
            return choice

        self.assertEqual(2, argmap.argmap(func, ["2"]))

    def test_choice_failure(self):
        def func(choice : {1, 2, 3}):
            return choice

        self.assertRaises(SystemExit, argmap.argmap, func, ["4"])

    def test_choice_kwarg(self):
        def func(choice : {1, 2, 3} = 4):
            return choice

        self.assertEqual(2, argmap.argmap(func, ["--choice", "2"]))

class TestVarArgs3(TestCase):
    def test_varargs_simple_type(self):
        def func(*args : float):
            return args

        self.assertSequenceEqual(argmap.argmap(func, ["1", "2", "3"]),
                                 [1, 2, 3])

    def test_varargs_tuple_type(self):
        def func(*args : (str, float)):
            return args

        self.assertSequenceEqual(argmap.argmap(func, ["1", "2", "3", "4"]),
                                 [("1", 2), ("3", 4)])

class DummyClass3(object):
    @classmethod
    def run(cls, a, b : argmap.none_or(int) = None):
        return a, b

class TestClassmethod3(TestCase):
    def test_classmethod(self):
        self.assertTupleEqual(argmap.argmap(DummyClass3.run,
                                            ["hello"]),
                              ("hello", None))

        self.assertTupleEqual(argmap.argmap(DummyClass3.run,
                                            ["hello", "--b", "42"]),
                              ("hello", 42))

if __name__ == "__main__":
    main()
